/*
 *  Name: canon
 *  Documentation:
 *    Canonicalize a plain text file.  The canonicalization converts
 *    text with a <CR><LF> sequence at the end of a line of text, no
 *    trailing spaces (0x20) or tabs (0x7e) on a line of text, and no
 *    trailing blank lines.  Also, leading byte-order-marks are trimmed.
 *
 *    Note: This text file canonicalization procedure is consistent
 *    with the ASCII NVT definition offered in Appendix B of RFC 5198.
 *
 *  Copyright (c) 2013, 2014, 2015, 2017, 2018 Russ Housley and Glen Barney.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice in the documentation and/or other materials provided with
 *     the distribution.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ''AS IS'' AND ANY
 *  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 *  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *  OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *  IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Version 1.0.0 -- 9 December 2013 -- Written by
 *    Russ Housley and Glen Barney.
 *  Version 1.0.1 -- 27 January 2014 -- Modified by Glen Barney and
 *    Russ Housley to remove trailing whitespace from the end of
 *    the file.
 *  Version 1.0.2 -- 10 March 2015 -- Modified by Glen Barney to
 *    add version and date defines; they used to be in the Makefile.
 *  Version 1.0.3 -- 21 March 2015 -- Modified by Russ Housley to
 *    exit after version or usage is provided.
 *  Version 1.0.4 -- 14 February 2017 -- Modified by Russ Housley
 *    to provide an error message when the regfile returns 1,
 *    which can happen if the input file does not exist.
 *  Version 1.0.5 -- 21 January 2018 -- Modified by Russ Housley and
 *    Glen Barney to trim any leading byte-order-marks and handle
 *    non-ASCII characters in the body of the file.
 */

#define	VER_REVISION	"1.0.5"
#define	VER_DATE   "2018-01-21"
#define	MAXLEN 65536

/* ASCII character codes */
#define A_NULL 0x00
#define A_LF   0x0a
#define A_FF   0x0c
#define A_CR   0x0d
#define A_SP   0x20
#define A_BOM1 0xef
#define A_BOM2 0xbb
#define A_BOM3 0xbf

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

/* display the usage lines on stderr */

void print_usage(void)
{
	fprintf(stderr, "\
canon %s (%s)\n\
Usage: canon [-h] [-v] [infile [outfile]]\n\
 -h --help        give this help\n\
 -v --version     display version number\n\
\n\
 If no infile is provided, then stdin is used.\n\
 If no outfile is provided, then stdout is used.\n",
 VER_REVISION, VER_DATE);
	}

/* display the version number on stderr */

void print_version(void)
{
	fprintf(stderr, "canon %s (%s)\n", VER_REVISION, VER_DATE);
	}

/* regfile confirms that the file is a normal file, not a special inode */

int regfile(char *filename)
{
	struct stat buf;
	if ((stat(filename, &buf) == 0) && S_ISREG(buf.st_mode)) return(1);
	else return(0);
	}

/* return number of bytes that are currently in the buffer */

int buflen(unsigned char *buf)
{
    int i,len=0;
    if(buf==NULL) return(0);
    for (i=0; buf[i]!=0 || i>=MAXLEN; i++) len++;
    if((i<0)||(i>=MAXLEN)) return(-1);
    return(len);
    }

/* write the buffer to the output file terminated with a CR and LF */

void writebuffer(FILE *outfile, unsigned char *buf)
{
    if(buf==NULL) return;
    int i=buflen(buf);
    if(i<0) return;
    if(i>0) fwrite(buf, i, 1, outfile);
    fprintf(outfile, "\r\n");
    }

/* remove spaces, tabs, and control characters from the end of a line */

void spacecut(unsigned char *buf)
{
    if(buf==NULL) return;
    int i=buflen(buf)-1;
    if((i<0)||(i>MAXLEN)) return;
    while((i>=0)&&(buf[i]!=A_FF)&&(buf[i]<=A_SP)) buf[i--]=A_NULL;
    }

/* return 1 if the array contains a byte-order-mark; otherwise return 0 */

int is_bom(unsigned char wt[3], int wtlen){
    if(wtlen!=3) return 0;
    if(wt[0]==A_BOM1 && wt[1]==A_BOM2 && wt[2]==A_BOM3) return 1;
    return 0;
    }

/* canonical steps through a file and cleans the lines */

void canonical(FILE *infile,FILE *outfile)
{
    unsigned char wt[3]={A_NULL,A_NULL,A_NULL};
    unsigned char c=0,p=0,*buffer;
	unsigned long k=0,l=0,wtlen=0;

	/* allocate memory */
	if((buffer=calloc(MAXLEN,1))==NULL){
		fprintf(stderr, "Unable to allocate memory");
		exit(1);
		}
    memset(buffer,A_NULL,MAXLEN);

    /* trim one or more leading byte-order-mark */
    if(!feof(infile)) { wt[0]=fgetc(infile); wtlen++; }
    if(!feof(infile)) { wt[1]=fgetc(infile); wtlen++; }
    if(!feof(infile)) { wt[2]=fgetc(infile); wtlen++; }
    while(is_bom(wt,wtlen)){
        wtlen=0;
        if(!feof(infile)) { wt[0]=fgetc(infile); wtlen++; }
        if(!feof(infile)) { wt[1]=fgetc(infile); wtlen++; }
        if(!feof(infile)) { wt[2]=fgetc(infile); wtlen++; }
        }
    if(wtlen==3) { ungetc(wt[2],infile); wtlen--; }
    if(wtlen==2) { ungetc(wt[1],infile); wtlen--; }
    if(wtlen==1) { ungetc(wt[0],infile); wtlen--; }
    
	/* scan file */
    c=fgetc(infile);
	while(!feof(infile)){
		/* special handling for CR-non-LF sequence */
		if((p==A_CR)&&(c!=A_LF)){ ungetc(c,infile); c=A_LF; }
		/* at end of line, clean and dump line */
		if(c==A_LF){
            spacecut(buffer);
            if(buffer[0]==0) l++;
            else{
                while(l>0){
                    fprintf(outfile,"\r\n");
                    l--;
                    }
                writebuffer(outfile,buffer);
                }
            memset(buffer,A_NULL,MAXLEN);
			k=0;
			}
		/* otherwise store character */
		else *(buffer+(k++))=c; 
		/* remember previous character and get the next one */
		p=c;
        c=fgetc(infile);
		/* protect against buffer overrun */
		if(k>=MAXLEN){
			fprintf(stderr, "Buffer overrun");
			exit(1);
			}
		}
	if(k>0){
		/* dump last line if partial */
        spacecut(buffer);
        if(buffer[0]){
            while(l>0){
                fprintf(outfile,"\r\n");
                l--;
                }
            writebuffer(outfile,buffer);
            }
        }

	/* release memory */
	free(buffer);
	}
			
/* main processing loop */

int main (int argc, char *argv[])
{
	int	argindex = 0;
	char	*infilename = NULL;
	char	*outfilename = NULL;
	FILE	*infile = NULL;
	FILE	*outfile = NULL;

	/* command line argument parsing */
	while (++argindex < argc){
		if (argv[argindex][0] == '-'){ /* an option */
			if ((strcmp(argv[argindex], "-h") == 0) || (strcmp(argv[argindex], "--help") == 0)){
			    print_usage();
			    exit(0);
			    }
			if ((strcmp(argv[argindex], "-v") == 0) || (strcmp(argv[argindex], "--version") == 0)){
			    print_version();
			    exit(0);
			    }
			}
		else{ /* not an option, so a filename was provided */
			if (infilename == NULL){
				infilename = argv[argindex];
                if (!regfile(infilename)){
                    fprintf(stderr, "Unable to open input file: %s\n", infilename);
                    exit(1);
                }
				infile = fopen(infilename, "r");
				if (infile == NULL){
					fprintf(stderr, "Unable to open input file: %s\n", infilename);
					exit(1);
					}
				}
			else{
				outfilename = argv[argindex];
				outfile = fopen(outfilename, "w");
				if (outfile == NULL){
					fprintf(stderr, "Unable to open output file: %s\n", outfilename);
					exit(1);
					}
				}
			}
		}

	/* perform canonicalization */
	if (infilename == NULL) infile = stdin;
	if (outfilename == NULL) outfile = stdout;

	canonical(infile, outfile);

	if (infilename != NULL) fclose(infile);
	if (outfilename != NULL) fclose(outfile);

	exit(0);
	}
